# -*- coding: iso-8859-1 -*-

# ArtHelp.py
#
# COM classes for the AR-Tool
#

from win32com.client           import GetObject, Dispatch
from win32com.server.exception import COMException
from os.path                   import exists, normcase, normpath, join
from os                        import listdir
from socket                    import gethostname
from subprocess                import Popen, PIPE, STDOUT
from tempfile                  import gettempdir
from shutil                    import copy
from distutils.sysconfig       import get_python_lib
from StringIO                  import StringIO
from textwrap                  import wrap

import winerror
import pythoncom
import win32api
import win32pipe
import win32con
import win32service
import win32serviceutil
import difflib
import re
import string
import sys
import xml.dom.minidom

__version__ = 330

False  = bool(0)
True   = bool(1)

# XML_Encoding
_XML_ENCODING = 'iso-8859-1'

# Constants for communication with the AR-Tool Utility Service (ArtUtilSv)

_ART_UTSV_NAME      = 'ArtUtilSv'
_ART_UTSV_DISP_NAME = 'AR-Tool Utility Service'
_ART_UTSV_PIPE      = '\\\\.\\pipe\\ARToolUtil'
_SV_NOT_FOUND       = -1

# Request
_RQ_TAG_NAME        = 'request'
_RC_ATTR_NAME       = 'rc'
_RC_GET_VERSION     = '1'
_RC_ASP_APP_RESTART = '2'

# answer codes
_AN_TAG_NAME        = 'answer'
_AC_ATTR_NAME       = 'ac'
_AC_OK              = '0'

# Constants for IIS administration

_APP_INPROC  = 0
_APP_OUTPROC = 1
_APP_POOLED  = 2

_IIS_PROTOCOL            = 'IIS:'
_IIS_SERVER              = 'LocalHost'
_IIS_WEB_SERVICE         = 'W3SVC'
_IIS_WEB_SERVER_ROOT     = 'ROOT'

_IIS_SERVER_ADSPATH           = '%s//%s' % (_IIS_PROTOCOL, _IIS_SERVER)
_IIS_WEB_SERVICE_ADSPATH      = '%s/%s'  % (_IIS_SERVER_ADSPATH, _IIS_WEB_SERVICE)

_IIS_COMPUTER_CLASS      = 'IIsComputer'
_IIS_WEBSERVICE_CLASS    = 'IIsWebService'
_IIS_WEBSERVER_CLASS     = 'IIsWebServer'
_IIS_WEBVIRTUALDIR_CLASS = 'IIsWebVirtualDir'
_IIS_WEBDIRECTORY_CLASS  = 'IIsWebDirectory'

_DEF_APPLICATION         = 'AR-Tool'
_DEF_VDIRKEY             = 'AR_Tool'
_DEF_PATH                = 'C:\\inetpub\wwwroot'

_PDF                     = ('pdf',)
_PDF_HTTPREDIRECT        = '*;*.pdf;pr_pdf.asp$Q, EXACT_DESTINATION'
_AUTOLOG                 = ('autolog',)
_NLS                     = ('nls',)

_WEB_DIR_NO_ACCESS       = (('conf',), ('include',), ('nls',), ('xsl',))
_WEB_DIR_READ_ONLY       = (('conf_ext',), ('css',), ('download',), ('img',), ('js',), ('schema',))

# Constants for ACL administration

_WNT_NAMESPACE        = 'WinNT://'
_WNT_COMPUTER_CLASS   = 'Computer'
_WNT_USER_CLASS       = 'User'
_WNT_GROUP_CLASS      = 'Group'

_WNT_SYSTEMINFO       = 'WinNTSystemInfo'

_AR_TOOL_USERS        = 'AR_Tool_Users'
_AR_TOOL_USERS_DESC   = 'Special Group for the AR-Tool Users'

_LOG_FILE_NAME        = 'acladmin_log.htm'
_STY_FILE_NAME        = 'setup.css'
_ANO_USER_PRE         = 'IUSR_'

ERROR_GROUP_NOT_FOUND = 2220

_BASE_ACL_OPTIONS = (('', 'r'), ('conf', 'f'), ('css', 'f'),)
_DBAS_ACL_OPTIONS = (('', 'f'),)
_LINK_ACL_OPTIONS = (('', 'f'),)
_UPLO_ACL_OPTIONS = (('', 'f'),)

# Constants for retrieving version information

_ARTHELP_VERSION_FILE = 'arthelp.version.txt'     # Name of the AR-Tool Helpers package version text file
_PYWIN32_VERSION_FILE = 'pywin32.version.txt'     # Name of the Win32 package version text file

# Utility functions and classes

def _CheckString(chars):

    RegExp = re.compile(r'[^\w-]')
    if RegExp.search(chars):
        raise COMException('Invalid characters in string: %s' % chars, winerror.ERROR_SERVICE_NOT_ACTIVE)

def _GetWin32ErrorCode(com_exc):
    
    hr, msg, exc, narg = com_exc
    if exc: hr = exc[-1]
    if winerror.HRESULT_FACILITY(hr) != winerror.FACILITY_WIN32: raise
    return winerror.SCODE_CODE(hr)

def _GetNodeText(xml_node):
    
    node_text = ''
    for node in xml_node.childNodes:
        if node.nodeType == node.TEXT_NODE: node_text = node_text + node.data
    return node_text

def _GetServiceStatus():

    try:
        return win32serviceutil.QueryServiceStatus(_ART_UTSV_NAME, None)[1]
    except:
        return _SV_NOT_FOUND

class XMLDocClass(xml.dom.minidom.Document):
    # Class for XML Document Objects, derived from xml.dom.minidom.Document
    # with additional utility functions

    def MakeNode(self, ParentNode=None, NodeName='root', NodeText=None, NodeAttrs={}):
        # Helper function to create and append a DOM node

        DomNode = self.createElement(NodeName)
        if NodeText:
            if type(NodeText) is unicode:
                NodeText = NodeText.encode(_XML_ENCODING)
            DomNode.appendChild(self.createTextNode(NodeText))
        if not ParentNode: ParentNode = self
        ParentNode.appendChild(DomNode)
        for Attr in NodeAttrs.items():
            DomNode.setAttribute(Attr[0], Attr[1])

        return DomNode

class CommArtUtilSvClass:

    def __init__(self, rc):

        # Initialize variables
        self.ac          = None
        self.msg         = ''
        self.request     = ''
        self.requestDoc  = XMLDocClass()
        self.requestNode = self.requestDoc.MakeNode(NodeName=_RQ_TAG_NAME, NodeAttrs={_RC_ATTR_NAME: rc})
        self.answer      = ''
        self.answerDoc   = None
        self.answerNode  = None      

    def CommServ(self):    

        if _GetServiceStatus() != win32service.SERVICE_RUNNING:
            #raise NameError('Test')
            raise COMException('AR-Tool Utility Service not active', winerror.ERROR_SERVICE_NOT_ACTIVE)

        # Send request to utility service and get answer        
        self.request = self.requestDoc.toxml()
        try:
            self.answer = win32pipe.CallNamedPipe(_ART_UTSV_PIPE, self.request, 1024, 0)
        except:
            raise COMException('Communication error', winerror.ERROR_BAD_PIPE)
        
        # Parse answer into a DOM object
        try:
            self.answerDoc = xml.dom.minidom.parseString(self.answer)
        except:
            raise COMException('Error when parsing the answer', winerror.ERROR_INVALID_DATA)
        
        # Get the DOM root object and check its tag name
        try:
            self.answerNode = self.answerDoc.documentElement
            if self.answerNode.nodeName != _AN_TAG_NAME:
                raise COMException('Top level tag is not named "%"' % _AN_TAG_NAME, winerror.ERROR_INVALID_DATA)
        except:
            raise COMException('No top level tag found', winerror.ERROR_INVALID_DATA)

        # Extract answer code
        self.ac = self.answerNode.getAttribute(_AC_ATTR_NAME)
        if not self.ac:
            raise COMException('Answer code missing or empty', winerror.ERROR_INVALID_DATA)

        # Extract error message if not ok
        if self.ac != _AC_OK:
            self.msg = _GetNodeText(self.answerNode)

        return          

# COM Classes

class PyVers:
    # A COM class for reading Python related versions
    # Obsolete as a COM class, equivalent function will be available as part of the NTAdmin class
    # COM relevant properties are still defined in order to allow for proper unregistering
    
    _register_me_ = False
    _reg_clsid_   = '{C03151D9-EC96-4540-BDE7-3B5F94717C90}'
    _reg_desc_    = 'Obsolete'
    _reg_progid_  = 'AR_Tool.PyVers'

    _public_methods_ = []
    _public_attrs_   = []
    _readonly_attrs_ = []

    def __init__(self):

        self.Version        = __version__
        
class NTAdmin:
    # A COM class for NT Administration
    # Supplies version information for Python related software for convenience, too
    
    _register_me_ = True
    _reg_clsid_   = '{6A538B80-A575-44BE-A264-C9D00B038F04}'
    _reg_desc_    = 'NT Administration'
    _reg_progid_  = 'AR_Tool.NTAdmin'

    _public_methods_ =  [
                        'Reload',
                        'SetAdminMode',
                        'FindVirtualDirs',
                        'GetDefApplication',
                        'GetDefVDirKey',
                        'CreateVirtualDirectory',
                        'DeleteVirtualDirectory',
                        'AcquireVirtualDirectory',
                        'CheckVirtualDirectory',
                        'SetVirtualDirectoryParameters',
                        'GetStruct',
                        'SetACLs',
                        'AspAppRestart',
                        'GetServiceStatus',
                        ]
    _readonly_attrs_ =  [
                        'Version',
                        'ServicePath',
                        'LogFilePath',
                        'Python_Version',
                        'Win32_Version',
                        'Helpers_Version',
                        'SV_NOT_FOUND',
                        'SV_STOPPED',
                        'SV_START_PENDING',
                        'SV_STOP_PENDING',
                        'SV_RUNNING',
                        ]
    _public_attrs_   =  _readonly_attrs_ + []

    def __init__(self):
        
        self.Version          = __version__

        self.SV_NOT_FOUND     = _SV_NOT_FOUND
        self.SV_STOPPED       = win32service.SERVICE_STOPPED
        self.SV_START_PENDING = win32service.SERVICE_START_PENDING
        self.SV_STOP_PENDING  = win32service.SERVICE_STOP_PENDING
        self.SV_RUNNING       = win32service.SERVICE_RUNNING
        self.LogFilePath      = ''
        self.ServerADsPaths   = []
        self.AdminMode        = False
        
        self.Python_Version   = '%s.%s.%s' % sys.version_info[:3]

        lib_path = get_python_lib()
        reg_expr = re.compile(r'\n')
        
        self.Win32_Version    = '000'
        try:
            version_file = file(join(lib_path, _PYWIN32_VERSION_FILE))
        except:
            version_file = None
        if version_file:
            tst = reg_expr.sub('', version_file.readline())
            version_file.close()
            if tst:
                self.Win32_Version = tst

        self.Helpers_Version  = '000'
        try:
            version_file = file(join(lib_path, _ARTHELP_VERSION_FILE))
        except:
            version_file = None
        if version_file:
            tst = reg_expr.sub('', version_file.readline())
            version_file.close()
            if tst:
                self.Helpers_Version = tst

    # Internal utilities
    
    def _GetVirtualDirs(self, ServerADsPath, include_root):
    # Return a list for all virtual directories of an individual web server
    # If "include_root" is True, the server root is included in the list at first position
    # The list element is a tuple (VDir, Root), where Root is true for the server root and false otherwise
        
        ServerRootADsPath = '%s/%s' % (ServerADsPath, _IIS_WEB_SERVER_ROOT)
        try:
            ServerRoot = GetObject(ServerRootADsPath)
        except pythoncom.com_error, details:
            rc = _GetWin32ErrorCode(details)
            if rc == winerror.ERROR_PATH_NOT_FOUND:
                raise COMException ('Server root "%s" not found' % ServerRootADsPath, winerror.ERROR_INVALID_DATA)
            else: raise
        if ServerRoot.Class != _IIS_WEBVIRTUALDIR_CLASS:
            raise COMException ('Unexpected class for server root: "%s"' % ServerRoot.Class, winerror.ERROR_INVALID_DATA)

        if (include_root):
            VDirTuples = [(ServerRoot, True)]
        else:
            VDirTuples = []
            
        enum = ServerRoot._NewEnum()
        if enum:
            while 1:
                item = enum.Next()
                if len(item) > 0:
                    ADSobj = item[0]
                    if ADSobj.Class == _IIS_WEBVIRTUALDIR_CLASS:
                        if ADSobj.AppRoot != ServerRoot.AppRoot: VDirTuples.append((ADSobj, False))
                else: break

        return VDirTuples

    def _FindVDirsByPath(self, PhysPath):
    # Return a list of all virtual directories (including the server roots) of the web service
    # (i.e. of all pertinent web servers), which point to a given physical path
    # The list consists of tuples as described for _GetVirtualDirs

        VDirTuples = []
        for ServerADsPath in self.ServerADsPaths:
            for VDir, Root  in self._GetVirtualDirs(ServerADsPath, include_root=True):
                if VDir.Path:
                    if normcase(normpath(VDir.Path)) == PhysPath:
                        VDirTuples.append((VDir, Root))
                
        return VDirTuples

    def _FindVDirsByAppl(self, Application):
    # Return a list of all virtual directories (including the server roots) of the web service
    # (i.e. of all pertinent web servers), which have the given application name
    # The list consists of tuples as described for _GetVirtualDirs

        Application = Application.upper()
        VDirTuples = []
        for ServerADsPath in self.ServerADsPaths:
            for VDir, Root in self._GetVirtualDirs(ServerADsPath, include_root=True):
                if VDir.AppFriendlyName:
                    if VDir.AppFriendlyName.upper() == Application:
                        VDirTuples.append((VDir, Root))
                
        return VDirTuples

    def _CreateWebDir(self, ParObj, RelPath):
    # Create a web directory (if not already existing) for a path relative to a parent object
    # The relative path is given as a tuple with the path keys

        for Key in RelPath:
            ADsPath = "%s/%s" % (ParObj.ADsPath, Key)
            try:
                ParObj = GetObject(ADsPath)
            except pythoncom.com_error, details:
                rc = _GetWin32ErrorCode(details)
                if rc == winerror.ERROR_PATH_NOT_FOUND:
                    ParObj = ParObj.Create(_IIS_WEBDIRECTORY_CLASS, Key)
                else: raise
            ParObj.SetInfo()

            if ParObj.Class != _IIS_WEBDIRECTORY_CLASS:
                raise COMException ('Unexpected class for web directory: "%s"' % ParObj.Class, winerror.ERROR_INVALID_DATA)

        return ParObj    

    def _CallCacls(self, LogFileObj, Path, options):
    # Calls the "cacls" utility program with the path and the options given
    # Results are written to the Log File Object

        if not Path: return 0
        if not exists(Path): return 0
        
        if options[0]: Path = normpath(join(Path, options[0]))
        LogFileObj.write('Path: %s Permission: %s<br>' % (Path, options[1]))

        args = ['cacls'] + [Path, '/t', '/e', '/g', '%s:%s' % (_AR_TOOL_USERS, options[1])]
        procObj = Popen(args, shell=True, stdin=PIPE, stdout=PIPE, stderr=STDOUT)
        # In batch mode, cacls asks "Are you sure (Y/N)?" and expects an "y" to be entered
        # So write an "y" to the stdin of the process
        procObj.stdin.write('y')
        procObj.stdin.close()
        # Now read the result as reported by cacls on stdout and write to log file
        # Note: The first characters read from stdout read "Are you sure (Y/B)?" and are discarded
        LogFileObj.write(procObj.stdout.read().split('?')[-1].replace('\r\n', '<br>'))
        RetCode = procObj.wait()
        LogFileObj.write('Returncode: %s<br><br>' % RetCode)
        
        return RetCode

    def _AddUserToGroup(self, AnoUserName=None):
    # Add a given NT user to the user group given by _AR_TOOL_USERS
    # The user must exist whereas the group is created, if it not exists
    # If the user ist not given, a default value is used
        
        ComputerName  = Dispatch(_WNT_SYSTEMINFO).ComputerName
        if not AnoUserName: AnoUserName = '%s%s' % (_ANO_USER_PRE, ComputerName)
        LocalComputer = GetObject('%s%s,%s' % (_WNT_NAMESPACE, ComputerName, _WNT_COMPUTER_CLASS))
        AnoUserObj    = LocalComputer.GetObject(_WNT_USER_CLASS, AnoUserName)
        AnoUserPath   = AnoUserObj.ADsPath

        try:
            ARToolUsers = LocalComputer.GetObject(_WNT_GROUP_CLASS, _AR_TOOL_USERS)
        except pythoncom.com_error, details:
            rc = _GetWin32ErrorCode(details)
            if rc == ERROR_GROUP_NOT_FOUND:
                ARToolUsers = LocalComputer.Create(_WNT_GROUP_CLASS, _AR_TOOL_USERS)
                ARToolUsers.Description = _AR_TOOL_USERS_DESC
                ARToolUsers.SetInfo()
            else: raise

        if not ARToolUsers.IsMember(AnoUserPath):
            ARToolUsers.Add(AnoUserPath)

    # External methods
    
    def Reload(self):
    # Reload the module containing this code
    # Does not require admin mode

        exec 'import %s'  % self.__module__
        exec 'reload(%s)' % self.__module__

    def SetAdminMode(self):
    # Set the admin mode for the NTAdmin object
    # Does not require admin mode

        try:
            Service = GetObject(_IIS_WEB_SERVICE_ADSPATH)
        except:
            raise COMException ('Service "%s" not found; IIS not installed?' % _IIS_WEB_SERVICE_ADSPATH, winerror.ERROR_INVALID_DATA)
        if Service.Class != _IIS_WEBSERVICE_CLASS:
            raise COMException ('Unexpected class for web service: %s' % Service.Class, winerror.ERROR_INVALID_DATA)

        enum = Service._NewEnum()
        self.ServerADsPaths = []
        if enum:
            while 1:
                item = enum.Next()
                if len(item) > 0:
                    ADSobj = item[0]
                    if ADSobj.Class == _IIS_WEBSERVER_CLASS: self.ServerADsPaths.append(ADSobj.ADsPath)
                else:
                    break
                
        self.AdminMode = True

        return        

    def FindVirtualDirs(self, PhysPath):
    # Returns an XML-Stream containing the pertinent virtual directories pointing to a given physical path
    # Requires admin mode

        if not self.AdminMode:
            raise COMException ('Administrator mode required', winerror.ERROR_INVALID_DATA)
    
        PhysPath = normcase(normpath(PhysPath))

        xmlDoc   = XMLDocClass()
        RootNode = xmlDoc.MakeNode()
        
        for VDir, Root in self._FindVDirsByPath(PhysPath):
            Attrs = {'Path': VDir.ADsPath, 'Appl': VDir.AppFriendlyName,
                     'Phys': normcase(normpath(VDir.Path)), 'Key': VDir.Name}
            if Root:
                Attrs['Root'] = '1'
            xmlDoc.MakeNode(ParentNode=RootNode, NodeName=VDir.Class, NodeAttrs=Attrs)

        return xmlDoc.toxml()

    def GetDefApplication(self):
    # Returns a default application name
    # Requires admin mode

        if not self.AdminMode:
            raise COMException ('Administrator mode required', winerror.ERROR_INVALID_DATA)
        
        DefAppl = _DEF_APPLICATION
        ind = 0
        while self._FindVDirsByAppl(DefAppl):
            ind = ind + 1
            DefAppl = '%s_%s' % (_DEF_APPLICATION, ind)
        return DefAppl
        
    def GetDefVDirKey(self):
    # Returns a default virtual directory name
    # Requires admin mode

        if not self.AdminMode:
            raise COMException ('Administrator mode required', winerror.ERROR_INVALID_DATA)

        DefVDK = _DEF_VDIRKEY
        ind = 0
        VDir = None
        while 1:
            for ServerADsPath in self.ServerADsPaths:
                try:
                    VDir = GetObject('%s/%s/%s' % (ServerADsPath, _IIS_WEB_SERVER_ROOT, DefVDK))
                    break
                except pythoncom.com_error, details:
                    rc = _GetWin32ErrorCode(details)
                    if rc == winerror.ERROR_PATH_NOT_FOUND: pass
                    else: raise

            if VDir: 
                VDir = None
                ind = ind + 1
                DefVDK = '%s_%s' % (_DEF_VDIRKEY, ind)
            else:
                break
            
        return DefVDK

    def CreateVirtualDirectory(self, ServerRootADsPath, VDirKey, PhysPath, Application):
    # Creates a virtual directory
    # Requires admin mode
    
        if not self.AdminMode:
            raise COMException ('Administrator mode required', winerror.ERROR_INVALID_DATA)

        if not ServerRootADsPath:
            raise COMException ('No server path given', winerror.ERROR_INVALID_DATA)

        if not VDirKey:
            raise COMException ('No virtual directory key given', winerror.ERROR_INVALID_DATA)
        _CheckString(VDirKey)
        if not PhysPath:
            raise COMException ('No physical path given', winerror.ERROR_INVALID_DATA)
        PhysPath = normcase(normpath(PhysPath))
        if not exists(PhysPath):
            raise COMException ('Physical path "%" does not exist' % PhysPath, winerror.ERROR_INVALID_DATA)
        if not Application:
            raise COMException ('No application name given', winerror.ERROR_INVALID_DATA)
        if self._FindVDirsByAppl(Application):
            raise COMException ('At least one virtual directory with application "%s" already exists' % Application, winerror.ERROR_INVALID_DATA)
        
        VDirADsPath       = '%s/%s' % (ServerRootADsPath, VDirKey)
        try:
            GetObject(VDirADsPath)
            raise COMException ('An object with path %s already exists' % VDirADsPath, winerror.ERROR_INVALID_DATA)
        except pythoncom.com_error, details:
            rc = _GetWin32ErrorCode(details)
            if rc == winerror.ERROR_PATH_NOT_FOUND: pass
            else: raise

        try:
            ServerRoot = GetObject(ServerRootADsPath)
        except pythoncom.com_error, details:
            rc = _GetWin32ErrorCode(details)
            if rc == winerror.ERROR_PATH_NOT_FOUND:
                raise COMException ('Server root "%s" not found' % ServerRootADsPath, winerror.ERROR_INVALID_DATA)
            else: raise
        if ServerRoot.Class != _IIS_WEBVIRTUALDIR_CLASS:
            raise COMException ('Unexpected class for server root: "%s"' % ServerRoot.Class, winerror.ERROR_INVALID_DATA)

        VDir = ServerRoot.Create(_IIS_WEBVIRTUALDIR_CLASS, VDirKey)
        VDir.Path            = PhysPath
        VDir.AccessRead      = True
        VDir.SetInfo()
        VDir.AppCreate2(_APP_POOLED)
        VDir.AppFriendlyName = Application
        VDir.SetInfo()
                
        for OtherVDir, Root in self._FindVDirsByPath(PhysPath):
            if OtherVDir.ADsPath != VDir.ADsPath:
                OtherVDir.Path = _DEF_PATH
                OtherVDir.SetInfo()
                
    def DeleteVirtualDirectory(self, ServerRootADsPath, VDirKey):
    # Deletes a virtual directory
    # Requires admin mode
    
        if not self.AdminMode:
            raise COMException ('Administrator mode required', winerror.ERROR_INVALID_DATA)

        if not ServerRootADsPath:
            raise COMException ('No server path given', winerror.ERROR_INVALID_DATA)
        if not VDirKey:
            raise COMException ('No virtual directory key given', winerror.ERROR_INVALID_DATA)
        _CheckString(VDirKey)

        VDirADsPath       = '%s/%s' % (ServerRootADsPath, VDirKey)
        try:
            GetObject(VDirADsPath)
        except pythoncom.com_error, details:
            rc = _GetWin32ErrorCode(details)
            if rc == winerror.ERROR_PATH_NOT_FOUND:
                raise COMException ('Virtual directory "%s" not found' % VDirADsPath, winerror.ERROR_INVALID_DATA)
                return
            else: raise

        try:
            ServerRoot = GetObject(ServerRootADsPath)
        except pythoncom.com_error, details:
            rc = _GetWin32ErrorCode(details)
            if rc == winerror.ERROR_PATH_NOT_FOUND:
                raise COMException ('Server root "%s" not found' % ServerRootADsPath, winerror.ERROR_INVALID_DATA)
            else: raise
        if ServerRoot.Class != _IIS_WEBVIRTUALDIR_CLASS:
            raise COMException ('Unexpected class for server root: "%s"' % ServerRoot.Class, winerror.ERROR_INVALID_DATA)

        ServerRoot.Delete(_IIS_WEBVIRTUALDIR_CLASS, VDirKey)
        ServerRoot.SetInfo()

    def AcquireVirtualDirectory(self, VDirADsPath, PhysPath, Application):
    # Acquires a virtual directory for the physical path given
    # Requires admin mode
    
        if not self.AdminMode:
            raise COMException ('Administrator mode required', winerror.ERROR_INVALID_DATA)

        if not VDirADsPath:
            raise COMException ('Virtual directory ADS Path not given', winerror.ERROR_INVALID_DATA)
        _CheckString(VDirADsPath.split('/')[-1])
        if not PhysPath:
            raise COMException ('No physical path given', winerror.ERROR_INVALID_DATA)
        PhysPath = normcase(normpath(PhysPath))
        if not exists(PhysPath):
            raise COMException ('Physical path "%" does not exist' % PhysPath, winerror.ERROR_PATH_NOT_FOUND)
        if not Application:
            raise COMException ('No application name given', winerror.ERROR_INVALID_DATA)

        try:
            VDir = GetObject(VDirADsPath)
        except pythoncom.com_error, details:
            rc = _GetWin32ErrorCode(details)
            if rc == winerror.ERROR_PATH_NOT_FOUND:
                raise COMException ('Virtual directory "%s" not found' % VDirADsPath, winerror.ERROR_INVALID_DATA)
            else: raise
        if VDir.Class != _IIS_WEBVIRTUALDIR_CLASS:
            raise COMException ('Unexpected class for virtual directory: "%s"' % VDir.Class, winerror.ERROR_INVALID_DATA)

        VDirTuples = self._FindVDirsByAppl(Application)
        if VDirTuples:
            if len(VDirTuples) == 1:
                OtherVDir, Root = VDirTuples[0]
                if OtherVDir.ADsPath != VDir.ADsPath:
                    raise COMException ('At least one virtual directory with application "%s" already exists' % Application, winerror.ERROR_INVALID_DATA)
            else:
                raise COMException ('At least one virtual directory with application "%s" already exists' % Application, winerror.ERROR_INVALID_DATA)

        VDir.Path            = PhysPath
        VDir.AccessRead      = True
        VDir.SetInfo()
        VDir.AppCreate2(_APP_POOLED)
        VDir.AppFriendlyName = Application
        VDir.SetInfo()

        for OtherVDir, Root in self._FindVDirsByPath(VDir.Path):
            if OtherVDir.ADsPath != VDir.ADsPath:
                OtherVDir.Path = _DEF_PATH
                OtherVDir.SetInfo()

    def CheckVirtualDirectory(self, VDirADsPath, PhysPath):
    # Check, if a virtual directory (1) exists and (2) points to the physical path given
    # Return the pertinent application name, if no error occurred
    # Requires admin mode
    
        if not self.AdminMode:
            raise COMException ('Administrator mode required', winerror.ERROR_INVALID_DATA)

        if not VDirADsPath:
            raise COMException ('Virtual directory ADS path not given', winerror.ERROR_INVALID_DATA)
        _CheckString(VDirADsPath.split('/')[-1])
        try:
            VDir = GetObject(VDirADsPath)
        except pythoncom.com_error, details:
            rc = _GetWin32ErrorCode(details)
            if rc == winerror.ERROR_PATH_NOT_FOUND:
                raise COMException ('Virtual directory "%s" not found' % VDirADsPath, winerror.ERROR_INVALID_DATA)
            else: raise
        if VDir.Class != _IIS_WEBVIRTUALDIR_CLASS:
            raise COMException ('Unexpected class for virtual directory: "%s"' % VDir.Class, winerror.ERROR_INVALID_DATA)

        if PhysPath:
            if normcase(normpath(VDir.Path)) != normcase(normpath(PhysPath)):
                raise COMException ('Unexpected physical path %s' % VDir.Path, winerror.ERROR_INVALID_DATA)

        return VDir.AppFriendlyName   

    def SetVirtualDirectoryParameters(self, VDirADsPath):
    # Set virtual directory parameters
    # Requires admin mode
    
        if not self.AdminMode:
            raise COMException ('Administrator mode required', winerror.ERROR_INVALID_DATA)

        if not VDirADsPath:
            raise COMException ('Virtual directory ADS Path not given', winerror.ERROR_INVALID_DATA)
        _CheckString(VDirADsPath.split('/')[-1])
        try:
            VDir = GetObject(VDirADsPath)
        except pythoncom.com_error, details:
            rc = _GetWin32ErrorCode(details)
            if rc == winerror.ERROR_PATH_NOT_FOUND:
                raise COMException ('Virtual directory "%s" not found' % VDirADsPath, winerror.ERROR_INVALID_DATA)
            else: raise
        if VDir.Class != _IIS_WEBVIRTUALDIR_CLASS:
            raise COMException ('Unexpected class for virtual directory: "%s"' % VDir.Class, winerror.ERROR_INVALID_DATA)

        VDir.AccessExecute         = False
        VDir.AccessRead            = True
        VDir.AccessWrite           = False
        VDir.AccessScript          = True
        VDir.AccessSource          = False
        VDir.ContentIndexed        = True
        VDir.DontLog               = False
        VDir.EnableDirBrowsing     = False
        VDir.EnableDefaultDoc      = True
        VDir.DefaultDoc            = 'pr_frame.asp'
        VDir.AuthAnonymous         = True
        VDir.AuthBasic             = False
        VDir.AuthNTLM              = False
        VDir.AnonymousPasswordSync = False

        self._AddUserToGroup(VDir.AnonymousUserName)

        WebDirs = []
        enum = VDir._NewEnum()
        if enum:
            while 1:
                item = enum.Next()
                if len(item) > 0:
                    ADSobj = item[0]
                    if ADSobj.Class == _IIS_WEBDIRECTORY_CLASS:
                        WebDirs.append(ADSobj)
                else: break

        for WebDir in WebDirs:
            VDir.Delete(_IIS_WEBDIRECTORY_CLASS, WebDir.Name)

        WebDir  = self._CreateWebDir(VDir, _PDF)
        WebDir.HttpRedirect = _PDF_HTTPREDIRECT
        WebDir.SetInfo()

        WebDir  = self._CreateWebDir(VDir, _AUTOLOG)
        WebDir.AuthAnonymous = False
        WebDir.AuthBasic     = False
        WebDir.AuthNTLM      = True
        WebDir.SetInfo()

        for RelPath in _WEB_DIR_NO_ACCESS:  
            WebDir  = self._CreateWebDir(VDir, RelPath)
            WebDir.AccessExecute = False
            WebDir.AccessRead    = False
            WebDir.AccessWrite   = False
            WebDir.AccessScript  = False
            WebDir.AccessSource  = False
            WebDir.SetInfo()

        WebDirReadOnly = list(_WEB_DIR_READ_ONLY)
            
        for Dir in listdir(join(VDir.Path, 'nls')):
            WebDirReadOnly.append(_NLS + (Dir,'help'))

        for RelPath in WebDirReadOnly:  
            WebDir  = self._CreateWebDir(VDir, RelPath)
            WebDir.AccessExecute = False
            WebDir.AccessRead    = True
            WebDir.AccessWrite   = False
            WebDir.AccessScript  = False
            WebDir.AccessSource  = False
            WebDir.SetInfo()

        VDir.SetInfo()

    def GetStruct(self, PhysPath=''):
    # Get the structure of the web:
    # web server
    #  web service
    #   first web server (root)
    #    virtual directory
    #    ...
    #   second web server (root)
    #    virtual directory
    #    ...
    #   .....
    # Return an XML-stream
    # If a physical path is given it will have a special markup in the result
    # Requires admin mode
    
        if not self.AdminMode:
            raise COMException ('Administrator mode required', winerror.ERROR_INVALID_DATA)

        if PhysPath: PhysPath = normcase(normpath(PhysPath))

        xmlDoc   = XMLDocClass()

        Attrs = {'Path': _IIS_SERVER_ADSPATH, 'Key': _IIS_SERVER, 'Desc': gethostname()}
        CompNode = xmlDoc.MakeNode(NodeName=_IIS_COMPUTER_CLASS, NodeAttrs=Attrs)
        
        Attrs = {'Path': _IIS_WEB_SERVICE_ADSPATH, 'Key': _IIS_WEB_SERVICE, 'Desc': 'Web Sites'}
        SvceNode = xmlDoc.MakeNode(ParentNode=CompNode, NodeName=_IIS_WEBSERVICE_CLASS, NodeAttrs=Attrs)

        indlist = range(len(self.ServerADsPaths))
        lastind = indlist[-1]
        for ind in indlist:
            
            ServerADsPath = self.ServerADsPaths[ind]
            try:
                Server = GetObject(ServerADsPath)
            except pythoncom.com_error, details:
                rc = _GetWin32ErrorCode(details)
                if rc == winerror.ERROR_PATH_NOT_FOUND:
                    raise COMException ('Server "%s" not found' % ServerADsPath, winerror.ERROR_INVALID_DATA)
                else: raise
            if Server.Class != _IIS_WEBSERVER_CLASS:
                raise COMException ('Unexpected class for server: "%s"' % Server.Class, winerror.ERROR_INVALID_DATA)
            Desc = Server.ServerComment

            ServerRootADsPath = '%s/%s' % (ServerADsPath, _IIS_WEB_SERVER_ROOT)
            try:
                ServerRoot = GetObject(ServerRootADsPath)
            except pythoncom.com_error, details:
                rc = _GetWin32ErrorCode(details)
                if rc == winerror.ERROR_PATH_NOT_FOUND:
                    raise COMException ('Server root "%s" not found' % ServerRootADsPath, winerror.ERROR_INVALID_DATA)
                else: raise
            if ServerRoot.Class != _IIS_WEBVIRTUALDIR_CLASS:
                raise COMException ('Unexpected class for server root: "%s"' % VDir.Class, winerror.ERROR_INVALID_DATA)
            Appl = ServerRoot.AppFriendlyName
            Phys = normcase(normpath(ServerRoot.Path))
            if Phys == PhysPath:
                Desc = '<b>%s<b>' % Desc
                Appl = '<b>%s<b>' % Appl
                Phys = '<b>%s<b>' % Phys
            Attrs = {'Path': ServerADsPath, 'Key': Server.ADsPath.split('/')[-1], 'VPath': ServerRootADsPath,
                     'Desc': Desc, 'Appl': Appl, 'Phys': Phys}
            ServNode = xmlDoc.MakeNode(ParentNode=SvceNode, NodeName=Server.Class, NodeAttrs=Attrs)

            for VDir, Root in self._GetVirtualDirs(ServerADsPath, include_root=False):
                Key  = VDir.ADsPath.split('/')[-1]
                Desc = Key
                Appl = VDir.AppFriendlyName
                Phys = normcase(normpath(VDir.Path))
                if Phys == PhysPath:
                    Desc = '<b>%s<b>' % Desc
                    Appl = '<b>%s<b>' % Appl
                    Phys = '<b>%s<b>' % Phys
                Attrs = {'Path': VDir.ADsPath, 'Key': Key, 'Desc': Desc, 'Appl': Appl, 'Phys': Phys}
                if ind < lastind:
                    Attrs['img'] = 'vertdot.gif'
                else:
                    Attrs['img'] = 'blank.gif'                    
                VdirNode = xmlDoc.MakeNode(ParentNode=ServNode, NodeName=VDir.Class, NodeAttrs=Attrs)

        return xmlDoc.toxml()
        
    def SetACLs(self, BasePath, DbasPath, LinkPath, UploPath, SetPath):
    # Sets the ACLs required
    # Requires admin mode
    
        if not self.AdminMode:
            raise COMException ('Administrator mode required', winerror.ERROR_INVALID_DATA)

        self.LogFilePath = gettempdir()
        StyFilePath      = join(normpath(SetPath), _STY_FILE_NAME)
        copy(StyFilePath, self.LogFilePath)
        self.LogFilePath = join(self.LogFilePath, _LOG_FILE_NAME)
        LogFileObj       = file(self.LogFilePath, 'w')
        LogFileObj.write('<html><head><title>Protocol</title>')
        LogFileObj.write('<link rel="stylesheet" TYPE="text/css" href="%s"></head><body><p>' % _STY_FILE_NAME)

        ErrFlag = False
        
        if not ErrFlag:
            for options in _BASE_ACL_OPTIONS:
                if self._CallCacls(LogFileObj, BasePath, options) != 0:
                    ErrFlag = True
                    break
            
        if not ErrFlag:
            for options in _DBAS_ACL_OPTIONS:
                if self._CallCacls(LogFileObj, DbasPath, options) != 0:
                    ErrFlag = True
                    break
                
        if not ErrFlag:
            for options in _LINK_ACL_OPTIONS:
                if self._CallCacls(LogFileObj, LinkPath, options) != 0:
                    ErrFlag = True
                    break
                
        if not ErrFlag:
            for options in _UPLO_ACL_OPTIONS:
                if self._CallCacls(LogFileObj, UploPath, options) != 0:
                    ErrFlag = True
                    break

        LogFileObj.write('</p></body></html>')
        LogFileObj.close()

        if ErrFlag:
            raise COMException ('Error when setting security options', winerror.ERROR_INVALID_DATA)

    def AspAppRestart(self, v_path):
    # Does not require admin mode, because it uses the AR-Tool Utility Service
    # "v_path" is expected as ADS path of the corresponding virtual
    # (1) relative to the ADS path of the pertinent web service
    # (2) full ADS path

        if not v_path:
            raise COMException('Virtual Directory relative path missing', winerror.ERROR_INVALID_DATA)

        if not v_path.upper().startswith(_IIS_WEB_SERVICE_ADSPATH.upper()):
            v_path = '%s/%s' % (_IIS_WEB_SERVICE_ADSPATH, v_path)

        CommArtUtilSv = CommArtUtilSvClass(_RC_ASP_APP_RESTART)
        CommArtUtilSv.requestNode.appendChild(CommArtUtilSv.requestDoc.createTextNode(v_path))
        CommArtUtilSv.CommServ()
        if CommArtUtilSv.ac != _AC_OK:
            raise COMException(CommArtUtilSv.msg, winerror.ERROR_INVALID_DATA)

        return        

    def GetServiceVersion(self):
    # Does not require admin mode, because it uses the AR-Tool Utility Service
    
        CommArtUtilSv = CommArtUtilSvClass(_RC_GET_VERSION)
        CommArtUtilSv.CommServ()
        if CommArtUtilSv.ac != _AC_OK:
            raise COMException(CommArtUtilSv.msg, winerror.ERROR_INVALID_DATA)

        return _GetNodeText(CommArtUtilSv.answerNode)

    def GetServiceStatus(self):
    # Does not require admin mode
    
        return _GetServiceStatus()

class RegExTest:
    # A COM class for using Python regular expression tests on strings and entire files
    
    _register_me_ = True
    _reg_clsid_  = '{8D6FC828-69FF-418F-B0F2-2F98677EAF25}'
    _reg_desc_   = 'Regular Expression'
    _reg_progid_ = 'AR_Tool.RegExTest'

    _public_methods_ = ['Reload', 'SetPattern', 'FileTest', 'StringTest']
    _public_attrs_   = ['Version']
    _readonly_attrs_ = ['Version']

    def __init__(self):
        
        self.Reg_Ex_List = []
        self.Version     = __version__

    def Reload(self):

        exec 'import %s'  % self.__module__
        exec 'reload(%s)' % self.__module__

    def SetPattern(self, Pattern):
        
        self.Reg_Ex_List = []
 
        PattItems = Pattern.split('&')

        for PattItem in PattItems:
            if not PattItem:
                raise COMException ('Internal error', winerror.ERROR_INTERNAL_ERROR)
            try:
                Reg_Ex = re.compile(PattItem, re.IGNORECASE)
            except:
                self.Reg_Ex_List = []
                return False
            self.Reg_Ex_List.append(Reg_Ex)

        return True
           
    def FileTest(self, FilePath):
        
        try:
            BinFile = file(FilePath, 'rb')
        except:
            raise COMException ('File "' + FilePath + '" could not be opened', winerror.ERROR_FILE_NOT_FOUND)
        TestRes = self.StringTest(BinFile.read())
        BinFile.close()
        return TestRes
    
    def StringTest(self, TestString):
        
        if not self.Reg_Ex_List:
            raise COMException ('No pattern defined', winerror.ERROR_INVALID_DATA)

        for Reg_Ex in self.Reg_Ex_List:
            if not Reg_Ex.search(TestString):
                return False
        return True

class CmpStrClass:
    # A COM class for comparing strings
    
    _register_me_ = True
    _reg_clsid_  = '{28AC135F-9958-4300-ACA4-F18DDEE747BD}'
    _reg_desc_   = 'Compare Strings'
    _reg_progid_ = 'AR_Tool.CmpStrClass'

    _public_methods_ = ['Reload', 'CmpStr']
    _public_attrs_   = ['Version']
    _readonly_attrs_ = ['Version']

    def __init__(self):
        
        self.Version  = __version__
        self.numlines = 5

    def Reload(self):
        exec 'import %s'  % self.__module__
        exec 'reload(%s)' % self.__module__
           
    def CmpStr(self, CmpStr1, CmpStr2, wrapping):
        
        if type(CmpStr1) is unicode:
            CmpStr1 = CmpStr1.encode(_XML_ENCODING)
        lines1 = []
        for l in StringIO(CmpStr1).readlines():
            lines1 = lines1 + wrap(l, wrapping)
        if type(CmpStr2) is unicode:
            CmpStr2 = CmpStr2.encode(_XML_ENCODING)
        lines2 = []
        for l in StringIO(CmpStr2).readlines():
            lines2 = lines2 + wrap(l, wrapping)
        return difflib.HtmlDiff().make_table(lines1, lines2,'1','2',context=False,numlines=self.numlines)

def Register():
    
    UseCommandLine('')

def Unregister():
    
    UseCommandLine('--unregister')

def UseCommandLine(cmd_opt=None):
    
    import win32com.server.register
    from types import ClassType
    if cmd_opt:
        while len(sys.argv) < 2:
            sys.argv.append('')
        sys.argv[0] = __file__
        sys.argv[1] = cmd_opt

    unregister = '--unregister' in sys.argv
    for item in globals().values():
        if type(item) is ClassType and item.__module__ == __name__:
            if hasattr(item, '_reg_clsid_') and hasattr(item, '_register_me_'):
                if unregister or item._register_me_:
                    win32com.server.register.UseCommandLine(item)
                
# Calling the script from the command line without argument will register the COM classes
# Calling the script from the command line with the "--unregister" argument will unregister them
if __name__ == '__main__':
    UseCommandLine()
