# -*- coding: iso-8859-1 -*-

# ArtUtilSv.py
#
# AR-Tool Utility Service
#

from xml.dom.minidom import parseString
from win32com.client import GetObject

import ar_tool.ArtHelp as ah
reload(ah)

import win32serviceutil
import win32service
import win32event
import win32pipe
import win32file
import win32evtlogutil
import pywintypes
import winerror
import servicemanager
import sys
import traceback

__version__ = 310

_WARNINGS = ()

class request_handler:

    def __init__(self, request):

        # Initialize variables
        self.err        = 0
        self.msg        = ''
        self.log        = ''
        self.rc         = ''
        self.answer     = ''
        self.answerDoc  = ah.XMLDocClass()
        self.answerNode = self.answerDoc.MakeNode(NodeName=ah._AN_TAG_NAME)

        # Save original request
        self.request = request

        # Parse request into a DOM object
        try:
            self.requestDoc = parseString(request)
        except:
            self.__err_set__(1, 'Error when parsing the request')
            return

        # Get the DOM root object and check its tag name
        try:
            self.requestNode = self.requestDoc.documentElement
            if self.requestNode.nodeName != ah._RQ_TAG_NAME:
                self.__err_set__(2, 'Top level tag is not named "%"' % ah._RQ_TAG_NAME)
                return
        except:
            self.__err_set__(3, 'No top level tag found')
            return

        # Extract request code
        self.rc = self.requestNode.getAttribute(ah._RC_ATTR_NAME)
        if not self.rc:
            self.__err_set__(4, 'Request code missing or empty')
            return
        
        # Check request code
        if not self.rc in (ah._RC_GET_VERSION, ah._RC_ASP_APP_RESTART):
            self.__err_set__(5, 'Unknown request code')
            return

    def __err_set__(self, err_code, err_text):
        self.err = err_code
        self.msg = err_text

    def handle_request(self):

        # Do individual request handling depending on request code by the pertinent handling functions
        if not self.err:
            if   self.rc == ah._RC_GET_VERSION:
                self.__get_version__()
            elif self.rc == ah._RC_ASP_APP_RESTART:
                self.__app_restart__()
            else:
                self.__err_set__(6, 'Internal error')

        if self.err:
            # Error Handling
            # Purge existing answerDoc, so individual handling functions need not care
            self.answerDoc.unlink()
            # Set up answerDoc for error
            self.answerNode = self.answerDoc.MakeNode(NodeName=ah._AN_TAG_NAME,
                                                      NodeAttrs={ah._AC_ATTR_NAME: str(self.err)},
                                                      NodeText=self.msg)
            # Create error log, if not a warning
            if not self.err in _WARNINGS:
                self.log = 'Invalid service request\n' + self.msg + '\n' + self.request
        else:
            # No error; answerDoc already set up properly by handling function
            # Set ok code
            self.answerNode.setAttribute(ah._AC_ATTR_NAME, ah._AC_OK)

        # In case of error (not warning) we will always have a log;
        # otherwise the handling function may have generated one
        # Log it to the event log
        if self.log:
            servicemanager.LogInfoMsg(self.log)
        # There should always be a properly set up answerDoc; serialize it and save it
        self.answer = self.answerDoc.toxml('iso-8859-1')

    def __get_version__(self):
        
        self.answerNode.appendChild(self.answerDoc.createTextNode(str(__version__)))

        return

    def __app_restart__(self):

        VirtDirADsPath = ah._GetNodeText(self.requestNode)
        if not VirtDirADsPath:
            self.__err_set__(200, 'Virtual directory relative path missing')
            return
        
        try:
            VirtDir = GetObject(VirtDirADsPath)
        except:
            self.__err_set__(202, 'Cannot get virtual directory object "%s"' % VirtDirADsPath)
            return

        if VirtDir.Class != ah._IIS_WEBVIRTUALDIR_CLASS:
            self.__err_set__(203, 'Unexpected class for virtual directory: "%s"' % VirtDir.Class)
            return

        try:
            VirtDir.AspAppRestart()
        except:
             self.__err_set__(204, 'Restart operation failed')

        return             
            
class ArtUtilSv(win32serviceutil.ServiceFramework):
    
    _register_me_      = True
    _svc_name_         = ah._ART_UTSV_NAME
    _svc_display_name_ = ah._ART_UTSV_DISP_NAME

    def __init__(self, args):
        win32serviceutil.ServiceFramework.__init__(self, args)
        # Create an event which we will use to wait on.
        # The "service stop" request will set this event.
        self.hWaitStop = win32event.CreateEvent(None, 0, 0, None)
        # We need to use overlapped IO for this, so we dont block when
        # waiting for a client to connect.  This is the only effective way
        # to handle either a client connection, or a service stop request.
        self.overlapped = pywintypes.OVERLAPPED()
        # And create an event to be used in the OVERLAPPED object.
        self.overlapped.hEvent = win32event.CreateEvent(None,0,0,None)
        # Finally initialize our Performance Monitor counters
    
    def SvcStop(self):
        # Before we do anything, tell the SCM we are starting the stop process.
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        # And set my event.
        win32event.SetEvent(self.hWaitStop)

    def SvcDoRun(self):
        # Log a "started" message to the event log.
        win32evtlogutil.AddSourceToRegistry(self._svc_name_, servicemanager.__file__)
        servicemanager.Initialize(self._svc_name_, servicemanager.__file__)
        servicemanager.LogMsg(
                servicemanager.EVENTLOG_INFORMATION_TYPE, 
                servicemanager.PYS_SERVICE_STARTED,
                (self._svc_name_, ''))
               
        # We create our named pipe.
        openMode = win32pipe.PIPE_ACCESS_DUPLEX | win32file.FILE_FLAG_OVERLAPPED
        pipeMode = win32pipe.PIPE_TYPE_MESSAGE
        
        # When running as a service, we must use special security for the pipe
        sa = pywintypes.SECURITY_ATTRIBUTES()
        # Say we do have a DACL, and it is empty
        # (ie, allow full access!)
        sa.SetSecurityDescriptorDacl ( 1, None, 0 )

        pipeHandle = win32pipe.CreateNamedPipe(ah._ART_UTSV_PIPE,
            openMode,
            pipeMode,
            win32pipe.PIPE_UNLIMITED_INSTANCES,
            0, 0, 6000, # default buffers, and 6 second timeout.
            sa)

        # Loop accepting and processing connections
        while 1:            
            hr = win32pipe.ConnectNamedPipe(pipeHandle, self.overlapped)
            
            if hr==winerror.ERROR_PIPE_CONNECTED:
                # Client is fast, and already connected - signal event
                win32event.SetEvent(self.overlapped.hEvent)
            # Wait for either a connection or a service stop request
            waitHandles = self.hWaitStop, self.overlapped.hEvent
            timeout = win32event.INFINITE
            rc = win32event.WaitForMultipleObjects(waitHandles, 0, timeout)
            if rc==win32event.WAIT_OBJECT_0:
                # Stop event
                break
            else:
                # Pipe event - read the data
                try:
                    hr, request = win32file.ReadFile(pipeHandle, 1024)
                    try:
                        rh = request_handler(str(request))
                        rh.handle_request()
                    except:
                        servicemanager.LogErrorMsg(traceback.format_exc())
                        win32pipe.DisconnectNamedPipe(pipeHandle)
                        continue
                    # Write answer back
                    win32file.WriteFile(pipeHandle, rh.answer)
                    # And disconnect from the client.
                    win32pipe.DisconnectNamedPipe(pipeHandle)
                    # Discard handler
                    rh = None
                except win32file.error:
                    # Client disconnected without sending data
                    # or before reading the response.
                    # Thats OK - just get the next connection
                    continue

        # Now log a "service stopped" message
        servicemanager.LogMsg(
                servicemanager.EVENTLOG_INFORMATION_TYPE, 
                servicemanager.PYS_SERVICE_STOPPED,
                (self._svc_name_, ''))

def Register():
    HandleCommandLine(['--startup=auto', 'install'])
    HandleCommandLine(['start'])

def Unregister():
    HandleCommandLine(['stop'])
    HandleCommandLine(['remove'])

def HandleCommandLine(cmd_opt=None):
    from types import ClassType

    for item in globals().values():
        if type(item) is ClassType and item.__module__ == __name__:
            if hasattr(item, '_svc_name_') and hasattr(item, '_register_me_'):
                if item._register_me_:
                    if cmd_opt:
                        install = 'install' in cmd_opt
                        win32serviceutil.HandleCommandLine(item, argv = [__file__] + cmd_opt)
                    else:
                        install = 'install' in sys.argv
                        win32serviceutil.HandleCommandLine(item)
                    if install:
                        hscm = win32service.OpenSCManager(None,None,win32service.SC_MANAGER_ALL_ACCESS)
                        try:
                            hs = win32serviceutil.SmartOpenService(hscm, 'ArtUtilSV', win32service.SERVICE_ALL_ACCESS)
                            try:
                                try:
                                    win32service.ChangeServiceConfig2(hs,
                                                                      win32service.SERVICE_CONFIG_FAILURE_ACTIONS,
                                                                      {'ResetPeriod': 0, 'RebootMsg': u'', 'Command': u'',
                                                                       'Actions':  ((win32service.SC_ACTION_RESTART, 0),
                                                                                    (win32service.SC_ACTION_RESTART, 0),
                                                                                    (win32service.SC_ACTION_RESTART, 0))
                                                                      }
                                                                     )
                                except NotImplementedError:
                                    # ChangeServiceConfig2 do only exist on OS > Win2K (not on NT) 
                                    pass

                            finally:
                                win32service.CloseServiceHandle(hs)
                        finally:
                            win32service.CloseServiceHandle(hscm)

# Calling the script from the command line will perform the actions as defined
# in "win32serviceutil.HandleCommandLine" depending on the current command line options, e.g.:
# "stop"
# "remove"
# "install"
# "start"
if __name__=='__main__':
    HandleCommandLine()



    
